"""
Implements tasks for testing purposes
"""
import datetime as dt
from typing import Optional, Tuple, Union

import numpy as np

from eolearn.core import EOPatch, EOTask, FeatureTypeSet
from eolearn.core.utils.common import is_discrete_type

from ..utils.types import Feature, TimePeriod


class DummyRasterFeatureTask(EOTask):
    """Creates a raster feature with random values"""

    def __init__(
        self,
        feature: Feature,
        shape: Tuple[int, ...],
        dtype: Union[np.dtype, type],
        min_value: float = 0,
        max_value: float = 1,
    ):
        """
        :param feature: A raster feature to be created.
        :param shape: Shape of the created feature array.
        :param dtype: A dtype of the feature.
        :param min_value: All feature values will be greater or equal to this value.
        :param max_value: If feature has a discrete dtype or max_value == min_value then all feature values will be
            lesser or equal to this value. Otherwise, all features will be strictly lesser to this value.
        """
        self.feature = self.parse_feature(feature, allowed_feature_types=FeatureTypeSet.RASTER_TYPES)
        self.shape = shape
        self.dtype = dtype
        self.min_value = min_value
        self.max_value = max_value

        feature_type, _ = self.feature
        if len(self.shape) != feature_type.ndim():
            raise ValueError(
                f"Feature {self.feature} should have {feature_type.ndim()}-dimensional shape but {self.shape} was given"
            )
        if feature_type.is_discrete() and not is_discrete_type(self.dtype):
            raise ValueError(f"Feature {self.feature} only supports discrete dtypes but {self.dtype} was given")

    def _get_random_raster(self, rng: np.random.Generator) -> np.ndarray:
        """Creates a raster array from given random generator."""
        if self.max_value == self.min_value:
            return np.full(self.shape, self.max_value, dtype=self.dtype)

        if is_discrete_type(self.dtype):
            return rng.integers(
                int(self.min_value), int(self.max_value), size=self.shape, dtype=self.dtype, endpoint=True
            )

        array = rng.random(size=self.shape)
        array = (self.max_value - self.min_value) * array + self.min_value
        return array.astype(self.dtype)

    def execute(self, eopatch: Optional[EOPatch] = None, seed: Optional[int] = None) -> EOPatch:
        """Generates a raster feature randomly with a given seed."""
        eopatch = eopatch or EOPatch()
        rng = np.random.default_rng(seed)

        eopatch[self.feature] = self._get_random_raster(rng)
        return eopatch


class DummyTimestampFeatureTask(EOTask):
    """Creates a timestamp feature with random timestamps"""

    def __init__(self, time_interval: TimePeriod, timestamp_num: int):
        """
        :param time_interval: A time interval `[start, end)` from where all timestamps will be generated.
        :param timestamp_num: Number of timestamp in the created timestamp feature.
        """
        self.time_interval = tuple(map(_ensure_datetime, time_interval))
        self.timestamp_num = timestamp_num

    def execute(self, eopatch: Optional[EOPatch] = None, seed: Optional[int] = None) -> EOPatch:
        """Generates timestamps randomly with a given seed."""
        eopatch = eopatch or EOPatch()
        rng = np.random.default_rng(seed)

        start_time, end_time = self.time_interval
        total_seconds = int((end_time - start_time).total_seconds())
        random_integers = rng.integers(total_seconds, size=self.timestamp_num)
        random_integers.sort()
        timestamps = [start_time + dt.timedelta(seconds=int(seconds)) for seconds in random_integers]

        eopatch.timestamp = timestamps
        return eopatch


def _ensure_datetime(timestamp: dt.date) -> dt.datetime:
    """Ensures that the given timestamp is a datetime and not a date object."""
    if isinstance(timestamp, dt.datetime):
        return timestamp
    return dt.datetime.fromordinal(timestamp.toordinal())



"""
Defines pipelines for ingesting and modifying BYOC collections.
"""
import datetime
import datetime as dt
import logging
from collections import defaultdict
from typing import Dict, List, Optional, Tuple, cast

import fs
import geopandas as gpd
import rasterio
from pydantic import Field, validator

from sentinelhub import CRS, BBox, DataCollection, SentinelHubBYOC
from sentinelhub.api.byoc import ByocCollection, ByocTile
from sentinelhub.geometry import Geometry

from eogrow.core.config import RawConfig

from ..core.pipeline import Pipeline
from ..utils.types import JsonDict
from ..utils.validators import (
    ensure_defined_together,
    ensure_exactly_one_defined,
    optional_field_validator,
    parse_data_collection,
)
from .export_maps import TIMESTAMP_FORMAT

LOGGER = logging.getLogger(__name__)


class IngestByocTilesPipeline(Pipeline):
    """Ingests .tiff files to a BYOC collection. Based on `ExportMapsPipeline` output for timeless features."""

    class Schema(Pipeline.Schema):
        byoc_tile_folder_key: str
        file_glob_pattern: str = Field("**/*.tiff", description="Pattern used to obtain the TIFF files to use")

        new_collection_name: Optional[str] = Field(description="Used for defining a new BYOC collection.")
        existing_collection: Optional[DataCollection] = Field(description="Used when updating and reingesting.")
        _parse_byoc_collection = optional_field_validator("existing_collection", parse_data_collection, pre=True)
        _ensure_exclusion = ensure_exactly_one_defined("new_collection_name", "existing_collection")

        is_temporal: bool = Field(
            False,
            description=(
                "If the BYOC is marked as temporal the pipeline will assume that the direct parent folder of a TIFF"
                " contains the sensing time, i.e. filesystem structure follows that used by `ExportMapsPipeline`."
                " Example of such a path is `UTM_32638/2019-01-04T07-48-37/BANDS_S2_L1C.tiff`."
            ),
        )
        sensing_time: Optional[datetime.datetime] = Field(
            None, description="Sensing time (ISO format) added to BYOC tiles. Only used for timeless collections."
        )

        @validator("sensing_time", pre=True)
        def _parse_sensing_time(cls, value: Optional[str], values: Dict[str, object]) -> Optional[datetime.datetime]:
            is_temporal = values["is_temporal"]
            if is_temporal and value is None:
                return None
            if not is_temporal and value is not None:
                return datetime.datetime.fromisoformat(value)
            raise AssertionError("Sensing time should be set for timeless BYOC collections.")

        cover_geometry_folder_key: Optional[str] = Field(description="Folder for supplying a custom cover geometry.")
        cover_geometry: Optional[str] = Field(description="Specifies a geometry file describing the cover geometry.")
        _ensure_cover_geometry = ensure_defined_together("cover_geometry_folder_key", "cover_geometry")

        reingest_existing: bool = Field(False, description="Whether to reingest or skip already ingested tiles.")

    config: Schema

    def __init__(self, config: Schema, raw_config: Optional[RawConfig] = None):
        super().__init__(config, raw_config)
        if not self.storage.is_on_aws():
            raise ValueError("Can only ingest for projects based on S3 storage.")
        project_folder = self.storage.config.project_folder
        self.bucket_name = project_folder.replace("s3://", "").split("/")[0]
        self._cover_geometry_df: Optional[gpd.GeoDataFrame] = None

    def get_byoc_collection(self, byoc_client: SentinelHubBYOC) -> JsonDict:
        """Obtains information about the existing collection or creates a new one."""
        if self.config.new_collection_name:
            collection_spec = ByocCollection(self.config.new_collection_name, s3_bucket=self.bucket_name)
            LOGGER.info("Creating new collection.")
            response = byoc_client.create_collection(collection_spec)
            LOGGER.info("Collection %s created.", response["id"])
        else:
            existing_collection = cast(DataCollection, self.config.existing_collection)  # due to validation
            LOGGER.info("Obtaining full information for collection id %s.", existing_collection.collection_id)
            response = byoc_client.get_collection(existing_collection)
        return response

    def get_tile_paths(self) -> Dict[str, List[str]]:
        """Collects the folders and filenames of .tiff files to be ingested.

        Paths are relative to bucket, not project.
        """
        tiff_paths = self._get_tiff_paths()

        folder_to_filename_map = defaultdict(list)
        for folder, filename in map(fs.path.split, tiff_paths):
            folder_to_filename_map[folder].append(filename)

        return dict(folder_to_filename_map)

    def _get_tiff_paths(self) -> List[str]:
        """Collects the paths to .tiff files to ingest. Paths are relative to bucket, not project."""
        folder = self.storage.get_folder(key=self.config.byoc_tile_folder_key)
        filesystem = self.storage.filesystem
        return [
            self._get_byoc_compliant_path(path)
            for path, _ in filesystem.glob(path=folder, pattern=self.config.file_glob_pattern)
        ]

    def _get_byoc_compliant_path(self, relative_path: str) -> str:
        """Transforms a project relative path to a bucket-name relative path that is required by BYOC."""
        absolute_path = fs.path.combine(self.config.storage.project_folder, relative_path)  # type: ignore[attr-defined]
        return absolute_path.replace(f"s3://{self.bucket_name}/", "")  # removes s3://<bucket-name>/

    def _prepare_tile(self, folder: str, tiff_paths: List[str]) -> Optional[ByocTile]:
        """Collects all required metainfo to create a BYOC tile for the given folder."""
        some_tiff = fs.path.join(folder, tiff_paths[0])
        cover_geometry = self._get_tile_cover_geometry(some_tiff)
        if cover_geometry.geometry.is_empty:
            return None

        sensing_time = self.config.sensing_time
        if self.config.is_temporal:
            # assumes file-system structure is equal to that of ExportMapsPipeline
            timestamp_folder = fs.path.split(folder)[1]
            sensing_time = dt.datetime.strptime(timestamp_folder, TIMESTAMP_FORMAT)

        return ByocTile(folder, cover_geometry=cover_geometry, sensing_time=sensing_time)

    def _get_tile_cover_geometry(self, tiff_path: str) -> Geometry:
        """Get geometry of the tile by intersecting the tiff geometry with the general cover geometry."""
        full_path = f"s3://{self.bucket_name}/" + tiff_path
        with rasterio.open(full_path) as tiff_data:
            tiff_bounds = tiff_data.bounds
            tiff_crs = CRS(tiff_data.crs.to_epsg())
        tiff_poly = BBox([tiff_bounds.left, tiff_bounds.bottom, tiff_bounds.right, tiff_bounds.top], tiff_crs).geometry

        cover_poly = self._get_cover_geometry(tiff_crs)
        final_poly = tiff_poly.intersection(cover_poly) if cover_poly else tiff_poly
        return Geometry(final_poly, crs=tiff_crs)

    def _get_cover_geometry(self, crs: CRS) -> Optional[Geometry]:
        """Lazy-loads the cover geometry of whole area, reprojecting (and combining) in desired CRS on call."""
        if self.config.cover_geometry is None or self.config.cover_geometry_folder_key is None:
            return None

        if self._cover_geometry_df is None:
            folder_path = self.storage.get_folder(self.config.cover_geometry_folder_key)
            file_path = fs.path.join(folder_path, self.config.cover_geometry)
            with self.storage.filesystem.openbin(file_path, "r") as file_handle:
                self._cover_geometry_df = gpd.read_file(file_handle, engine=self.storage.config.geopandas_backend)

        return self._cover_geometry_df.to_crs(crs.pyproj_crs()).unary_union

    def run_procedure(self) -> Tuple[List[str], List[str]]:
        """Runs the procedure.

        1. Creates or loads the collection,
        2. Checks existing tiles,
        3. Creates new tiles and skips/reingests existing ones.
        """
        byoc_client = SentinelHubBYOC(config=self.storage.sh_config)
        byoc_collection = self.get_byoc_collection(byoc_client)

        existing_tiles = {fs.path.dirname(tile["path"]): tile["id"] for tile in byoc_client.iter_tiles(byoc_collection)}

        for tile_folder, tiff_paths in self.get_tile_paths().items():
            if tile_folder in existing_tiles:
                # reingest or skip existing
                if self.config.reingest_existing:
                    response = byoc_client.reingest_tile(byoc_collection, existing_tiles[tile_folder])
                    LOGGER.info(f"Reingested tile for {tile_folder}{f'with response: {response}' if response else ''}.")
                else:
                    LOGGER.info(f"Tile {tile_folder} already exists, skipping.")
            else:
                tile = self._prepare_tile(tile_folder, tiff_paths)
                if tile is None:
                    LOGGER.info(f"Intersection of tile {tile_folder} with cover geometry is empty, skipping.")
                    continue
                response = byoc_client.create_tile(byoc_collection, tile)
                if "errors" in response:
                    LOGGER.info(f"Creation of tile {tile_folder} failed with response: {response}.")
                LOGGER.info(f"Created tile for {tile_folder}.")

        return [], []



"""
Module defining a base pipeline schema and custom fields

For each pipeline a separate schema has to be defined which inherits from PipelineSchema. Such schema should be placed
as an internal class of the implemented pipeline class
"""
from inspect import isclass
from typing import List, Optional, Type

from pydantic import BaseModel, Field
from pydantic.fields import ModelField

from ..utils.types import BoolOrAuto, ImportPath, Path
from ..utils.validators import field_validator, validate_manager
from .base import EOGrowObject

BaseSchema = EOGrowObject.Schema


class ManagerSchema(BaseSchema):
    """A basic schema for managers, to be used as a parent class for defining manager schemas"""

    manager: Optional[ImportPath] = Field(description="An import path to this specific manager.")


class PipelineSchema(BaseSchema):
    """Base schema of the Pipeline class."""

    pipeline: Optional[ImportPath] = Field(description="Import path to an implementation of Pipeline class.")

    storage: ManagerSchema = Field(description="A schema of an implementation of StorageManager class")
    validate_storage = field_validator("storage", validate_manager, pre=True)

    area: ManagerSchema = Field(description="A schema of an implementation of AreaManager class")
    validate_area = field_validator("area", validate_manager, pre=True)

    eopatch: ManagerSchema = Field(description="A schema of an implementation of EOPatchManager class")
    validate_eopatch = field_validator("eopatch", validate_manager, pre=True)

    logging: ManagerSchema = Field(description="A schema of an implementation of LoggingManager class")
    validate_logging = field_validator("logging", validate_manager, pre=True)

    workers: int = Field(
        1, description="Number of workers for parallel execution of workflows. Parameter does not affect ray clusters."
    )
    use_ray: BoolOrAuto = Field(
        "auto",
        description=(
            "Whether to run the pipeline locally or using a (local or remote) ray cluster. When using `auto` the"
            " pipeline checks if it can connect to a cluster, and if none are available runs locally."
        ),
    )

    patch_list: Optional[List[int]] = Field(
        description="A list of EOPatch indices for which the pipeline should be executed"
    )
    input_patch_file: Optional[Path] = Field(description="File path to a file with input EOPatch names")
    skip_existing: bool = Field(
        False,
        description=(
            "Whether or not to skip already processed patches. In order to use this functionality a pipeline "
            "must implement the `filter_patch_list` method."
        ),
    )


def build_schema_template(
    schema: Type[BaseModel],
    required_only: bool = False,
    pipeline_import_path: Optional[str] = None,
    add_descriptions: bool = False,
) -> dict:
    rec_flags: dict = dict(required_only=required_only, add_descriptions=add_descriptions)  # type is needed

    template: dict = {}
    for name, field in schema.__fields__.items():
        if required_only and not field.required:
            continue

        description = field.field_info.description if add_descriptions else None

        if name == "pipeline" and pipeline_import_path:
            template[name] = pipeline_import_path
        elif isclass(field.type_) and issubclass(field.type_, BaseModel):
            # Contains a subschema in the nesting
            if isclass(field.outer_type_) and issubclass(field.outer_type_, BaseModel):
                template[name] = build_schema_template(field.type_, **rec_flags)
                if description:
                    template[name]["<< description >>"] = description
            else:
                template[name] = {
                    "<< type >>": repr(field._type_display()),
                    "<< nested schema >>": str(field.type_),
                    "<< sub-template >>": build_schema_template(field.type_, **rec_flags),
                }
        else:
            template[name] = _field_description(field, description)

    return template


def _field_description(field: ModelField, description: Optional[str]) -> str:
    description = f" // {description}" if description else ""
    field_type = repr(field._type_display())
    default = repr(field.default) + " : " if field.default else ""
    return f"<< {default}{field_type}{description} >>"



"""
Module implementing command line interface for `eo-grow`
"""
import json
import os
import re
import subprocess
from typing import Optional, Tuple

import click

from .core.config import collect_configs_from_path, decode_config_list, encode_config_list, interpret_config_from_dict
from .core.schemas import build_schema_template
from .pipelines.testing import TestPipeline
from .utils.general import jsonify
from .utils.meta import collect_schema, import_object, load_pipeline_class

variables_option = click.option(
    "-v",
    "--variable",
    "cli_variables",
    multiple=True,
    type=str,
    help='Specify variables to use in config (overriding present config values). Must be of form `"name:value"`',
)
test_patches_option = click.option(
    "-t",
    "--test",
    "test_patches",
    multiple=True,
    type=int,
    help=(
        "One or more indices of EOPatches. If given, a pipeline will run only for those EOPatches. "
        "Example: -t 0 -t 42 will run the pipeline for EOPatches with indices 0 and 42."
    ),
)


class EOGrowCli:
    """A command line interface class for `eo-grow`

    It is designed to be transferable to other packages based on `eo-grow`.

    Note: It looks that click doesn't allow implementing CLI in a proper class structure, therefore here we are using
    class variable to store the list of packages (https://github.com/pallets/click/issues/601).
    """

    _command_namespace = "eogrow"

    def __init__(self, command_namespace: Optional[str] = None):
        """
        :param command_namespace: A namespace for calling CLI, e.g. command_namespace='eogrow' if you call commands
            "eogrow config.json".
        """
        if command_namespace:
            EOGrowCli._command_namespace = command_namespace

    @staticmethod
    @click.command()
    @click.argument("config_filename_or_string", type=click.Path())
    @variables_option
    @test_patches_option
    @click.option(
        "-e",
        "--encoding",
        "encoding",
        is_flag=True,
        type=bool,
        help="The string passed to method is treated as an encoded config instead of filename.",
    )
    def main(
        config_filename_or_string: str, cli_variables: Tuple[str], test_patches: Tuple[int], encoding: bool
    ) -> None:
        """Execute eo-grow pipeline using CLI.

        \b
        Example:
            eogrow config_files/config.json
        """
        if encoding:
            raw_configs = decode_config_list(config_filename_or_string)
        else:
            raw_configs = collect_configs_from_path(config_filename_or_string)

        cli_variable_mapping = dict(_parse_cli_variable(cli_var) for cli_var in cli_variables)
        pipelines = []
        for raw_config in raw_configs:
            config = interpret_config_from_dict(raw_config, cli_variable_mapping)
            if test_patches:
                config["patch_list"] = list(test_patches)

            pipelines.append(load_pipeline_class(config).from_raw_config(config))

        for pipeline in pipelines:
            pipeline.run()

    @staticmethod
    @click.command()
    @click.argument("cluster_yaml", type=click.Path())
    @click.argument("config_filename", type=click.Path())
    @click.option(
        "--start", "start_cluster", is_flag=True, type=bool, help="Starts the cluster if it is not currently running."
    )
    @click.option(
        "--stop",
        "stop_cluster",
        is_flag=True,
        type=bool,
        help=(
            "Stops the cluster if after running the pipeline. In order for this to work got to AWS console "
            "-> IAM -> Roles -> select ray-autoscaler-v1 role and attach IAMReadOnlyAccess policy."
        ),
    )
    @click.option(
        "--screen",
        "use_screen",
        is_flag=True,
        type=bool,
        help=(
            "Run the cluster in a detached mode using screen software. Use Ctrl+A+D to detach any time, "
            "even when running a pipeline or a Jupyter notebook. Use Ctrl+D to terminate the remote screen."
        ),
    )
    @click.option(
        "--tmux",
        "use_tmux",
        is_flag=True,
        type=bool,
        help="Run the cluster in a detached mode using tmux software. Use Ctrl+B and d to detach.",
    )
    @variables_option
    @test_patches_option
    def ray(
        config_filename: str,
        cluster_yaml: str,
        start_cluster: bool,
        stop_cluster: bool,
        use_screen: bool,
        use_tmux: bool,
        cli_variables: Tuple[str],
        test_patches: Tuple[int],
    ) -> None:
        """Command for running an eo-grow pipeline on a remote Ray cluster of AWS EC2 instances

        \b
        Example:
            eogrow-ray cluster.yaml config_files/config.json
        """
        if stop_cluster and (use_screen or use_tmux):
            raise NotImplementedError("It is not clear how to combine stop flag with either screen or tmux flag")

        configs = collect_configs_from_path(config_filename)
        encoded_configs = encode_config_list(configs)
        cmd = (
            f"{EOGrowCli._command_namespace} -e {encoded_configs}"
            + "".join(f' -v "{cli_var_spec}"' for cli_var_spec in cli_variables)
            + "".join(f" -t {patch_index}" for patch_index in test_patches)
            + ("; " if stop_cluster else "")  # Otherwise, ray will incorrectly prepare a command for stopping a cluster
        )
        flag_info = [("start", start_cluster), ("stop", stop_cluster), ("screen", use_screen), ("tmux", use_tmux)]
        exec_flags = " ".join(f"--{flag_name}" for flag_name, use_flag in flag_info if use_flag)

        subprocess.call(f"ray exec {exec_flags} {cluster_yaml} '{cmd}'", shell=True)

    @staticmethod
    @click.command()
    @click.argument("import_path", type=str)
    @click.argument("template_path", type=click.Path(), required=False)
    @click.option(
        "-f",
        "--force",
        "force_override",
        is_flag=True,
        type=bool,
        help=(
            "In case a template path is provided and a file in the path already exists this flag is used to force "
            "override it."
        ),
    )
    @click.option(
        "--template-format",
        "template_format",
        type=click.Choice(["minimal", "open-api"], case_sensitive=False),
        help="Specifies which template format to use. The default is `minimal`",
        default="minimal",
    )
    @click.option(
        "--required-only",
        "required_only",
        is_flag=True,
        type=bool,
        help="If provided it will only include required fields in the template. Only for `minimal` template format",
    )
    @click.option(
        "--add-descriptions",
        "add_descriptions",
        is_flag=True,
        type=bool,
        help="Adds descriptions to template. Only for `minimal` template format",
    )
    def make_template(
        import_path: str,
        template_path: Optional[str],
        force_override: bool,
        template_format: str,
        required_only: bool,
        add_descriptions: bool,
    ) -> None:
        """Command for creating a config template for an eo-grow pipeline

        \b
        Examples:
            - save template to file:
                eogrow-template eogrow.pipelines.download.DownloadPipeline config_files/download.json
            - print template to command line:
                eogrow-template eogrow.pipelines.download.DownloadPipeline
        """
        if not force_override and template_path and os.path.isfile(template_path):
            raise FileExistsError(f"File {template_path} already exists. You can use -f to force override it.")

        object_with_schema = import_object(import_path)
        schema = collect_schema(object_with_schema)

        if template_format == "open-api":
            template = schema.schema()
        else:
            template = build_schema_template(
                schema,
                pipeline_import_path=import_path,
                required_only=required_only,
                add_descriptions=add_descriptions,
            )

        if template_path:
            with open(template_path, "w") as file:
                json.dump(template, file, indent=2, default=jsonify)
        else:
            click.echo(json.dumps(template, indent=2, default=jsonify))

    @staticmethod
    @click.command()
    @click.argument("config_filename", type=click.Path())
    def validate_config(config_filename: str) -> None:
        """Validate config without running a pipeline.

        \b
        Example:
            eogrow-validate config_files/config.json
        """
        for config in collect_configs_from_path(config_filename):
            raw_config = interpret_config_from_dict(config)
            load_pipeline_class(config).Schema.parse_obj(raw_config)

        click.echo("Config validation succeeded!")

    @staticmethod
    @click.command()
    @click.argument("config_filename", type=click.Path())
    def run_test_pipeline(config_filename: str) -> None:
        """Runs a test pipeline that only makes sure the managers work correctly. This can be used to select best
        area manager parameters.

        \b
        Example:
            eogrow-test any_pipeline_config.json
        """
        for crude_config in collect_configs_from_path(config_filename):
            raw_config = interpret_config_from_dict(crude_config)
            pipeline = TestPipeline.with_defaults(raw_config)
            pipeline.run()


def _parse_cli_variable(mapping_str: str) -> Tuple[str, str]:
    """Checks that the input is of shape `name:value` and then splits it into a tuple"""
    match = re.match(r"(?P<name>.+?):(?P<value>.+)", mapping_str)
    if match is None:
        raise ValueError(f'CLI variable input {mapping_str} is not of form `"name:value"`')
    parsed = match.groupdict()
    return parsed["name"], parsed["value"]



"""
Module implementing utilities for unit testing pipeline results
"""
import functools
import json
import os
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, cast

import fs
import numpy as np
import pandas as pd
import rasterio
import shapely.ops
from deepdiff import DeepDiff
from fs.base import FS

from eolearn.core import EOPatch, FeatureType
from sentinelhub import BBox

from ..core.config import collect_configs_from_path, interpret_config_from_dict
from ..core.pipeline import Pipeline
from ..utils.meta import load_pipeline_class
from ..utils.types import JsonDict


class ContentTester:
    """Utility used for testing pipeline results

    Results of a pipeline is usually a folder containing multiple EOPatches, each containing various features. This
    utility aggregates all folder content into some basic statistics.

    * Every time you initialize this class statistics will be calculated
    * Statistics can be saved into a JSON file (it's human-readable)
    * Statistics can be compared with the one saved in a file from the previous run.

    If statistics match there is a good chance that the pipeline produced exactly the same results as before. Otherwise,
    this utility will let you know which statistics does not match
    """

    _STATS_OPERATIONS: Dict[str, Callable] = {
        "min": np.min,
        "max": np.max,
        "mean": np.mean,
        "median": np.median,
    }

    def __init__(
        self,
        filesystem: FS,
        main_folder: str,
        decimals: int = 5,
        unique_values_limit: int = 8,
        histogram_bin_num: int = 8,
        max_random_values: int = 5,
    ):
        """
        :param filesystem: A filesystem containing project data
        :param main_folder: A folder path on the filesystem where results are saved
        :param decimals: Number of decimals to which values will be rounded
        :param unique_values_limit: If a raster has at most this many unique values then statistics will show all
            unique values with their counts. Otherwise, multiple statistical properties will be calculated for
            the values.
        :param histogram_bin_num: Number of bins in a histogram for statistics. The histogram will be calculated only
            if number of unique values is higher than `unique_values_limit`.
        :param max_random_values: Number of values that will be randomly sampled from an array for statistics. This
            will happen only if the array contains at least `2` different unique values.
        """
        self.filesystem = filesystem
        self.main_folder = main_folder
        self.decimals = decimals
        self.unique_values_limit = unique_values_limit
        self.histogram_bin_num = histogram_bin_num
        self.max_random_values = max_random_values

        if not filesystem.isdir(self.main_folder):
            raise ValueError(f"Folder {self.main_folder} does not exist on filesystem {self.filesystem}")

        self.stats = self._calculate_stats()

    def compare(self, filename: str) -> DeepDiff:
        """Compares statistics of given folder content with statistics saved in a given file

        :param filename: A JSON filename (with file path) where expected statistics is saved
        :return: A dictionary report about differences between expected and actual content
        """
        with open(filename, "r") as file:
            expected_stats = json.load(file)

        return DeepDiff(expected_stats, self.stats)

    def save(self, filename: str) -> None:
        """Saves statistics of given folder content into a JSON file

        :param filename: A JSON filename (with file path) where statistics should be saved
        """
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        with open(filename, "w") as file:
            json.dump(self.stats, file, indent=2, sort_keys=True)

    def _calculate_stats(self, folder: Optional[str] = None) -> JsonDict:
        """Calculates statistics of given folder and it's content"""
        stats: Dict[str, object] = {}
        if folder is None:
            folder = self.main_folder

        for content in self.filesystem.listdir(folder):
            content_path = fs.path.combine(folder, content)

            if self.filesystem.isdir(content_path):
                eopatch = EOPatch.load(content_path, filesystem=self.filesystem)

                if eopatch.bbox:
                    stats[content] = self._calculate_eopatch_stats(eopatch)
                else:  # Probably it is not an EOPatch folder
                    stats[content] = self._calculate_stats(folder=content_path)
            elif content_path.endswith("tiff"):
                stats[content] = self._calculate_tiff_stats(content_path)
            elif content_path.endswith(".npy"):
                stats[content] = self._calculate_numpy_file_stats(content_path)
            elif content_path.endswith(".aux.xml"):
                pass
            else:
                stats[content] = None

        return stats

    def _calculate_eopatch_stats(self, eopatch: EOPatch) -> JsonDict:
        """Calculates statistics of given EOPatch and it's content"""
        stats: Dict[str, object] = {}

        for feature_type, feature_set in eopatch.get_features().items():
            feature_type_name = feature_type.value

            if feature_type is FeatureType.BBOX:
                stats[feature_type_name] = repr(eopatch.bbox)

            elif feature_type is FeatureType.TIMESTAMP:
                stats[feature_type_name] = [time.isoformat() for time in eopatch.timestamp]

            else:
                feature_set = cast(Set[str], feature_set)
                feature_stats_dict = {}

                if feature_type.is_raster():
                    calculation_method: Callable = self._calculate_numpy_stats
                elif feature_type.is_vector():
                    calculation_method = self._calculate_vector_stats
                else:  # Only FeatureType.META_INFO remains
                    calculation_method = str

                for feature_name in feature_set:
                    feature_stats_dict[feature_name] = calculation_method(eopatch[feature_type][feature_name])

                stats[feature_type_name] = feature_stats_dict

        return stats

    def _calculate_numpy_stats(self, raster: np.ndarray) -> JsonDict:
        """Calculates statistics over a raster numpy array"""
        stats: JsonDict = {"array_shape": list(raster.shape), "dtype": str(raster.dtype)}
        if raster.dtype == object or raster.dtype.kind == "U":
            return stats

        unique_values = np.unique(raster)

        if unique_values.size <= self.unique_values_limit:
            values, counts = np.unique(raster, return_counts=True)
            stats["values"] = [
                {"value": self._prepare_value(value), "count": int(count)} for value, count in zip(values, counts)
            ]

        else:
            number_values = raster[~np.isnan(raster)]
            finite_values = number_values[np.isfinite(number_values)]

            stats["counts"] = {
                "nan": raster.size - number_values.size,
                "infinite": number_values.size - finite_values.size,
            }
            stats["basic_stats"] = {
                name: self._prepare_value(operation(finite_values))
                for name, operation in self._STATS_OPERATIONS.items()
            }

            stats["subsample_basic_stats"] = self._calculate_subsample_stats(finite_values)

            counts, edges = np.histogram(finite_values, bins=self.histogram_bin_num)
            stats["histogram"] = {
                "counts": counts.astype(int).tolist(),
                "edges": list(map(self._prepare_value, edges)),
            }

        if unique_values.size > 1:
            stats["random_values"] = self._get_random_stats(raster, unique_values)

        return stats

    def _calculate_tiff_stats(self, tiff_filename: str) -> JsonDict:
        """Calculates statistics over a .tiff image"""
        with self.filesystem.openbin(tiff_filename, "r") as file_handle:
            with rasterio.open(file_handle) as tiff:
                image = tiff.read()
                mask = tiff.dataset_mask()

        return {
            "image": self._calculate_numpy_stats(image),
            "mask": self._calculate_numpy_stats(mask),
        }

    def _calculate_numpy_file_stats(self, numpy_filename: str) -> JsonDict:
        """Calculates statistics over a .npy file containing a numpy array"""
        with self.filesystem.openbin(numpy_filename, "r") as file_handle:
            raster = np.load(file_handle, allow_pickle=True)

        return self._calculate_numpy_stats(raster)

    def _calculate_vector_stats(self, dataframe: pd.DataFrame) -> JsonDict:
        """Calculates statistics over a vector GeoDataFrame"""  # TODO: add more statistical properties
        rounder = functools.partial(_round_point_coords, decimals=self.decimals)
        dataframe["geometry"] = dataframe["geometry"].apply(lambda geometry: shapely.ops.transform(rounder, geometry))

        stats = {
            "columns": list(dataframe),
            "row_count": len(dataframe.index),
            "crs": str(dataframe.crs),
        }

        if len(dataframe.index):
            stats["first_row"] = list(map(str, dataframe.iloc[0]))

        return stats

    def _calculate_subsample_stats(self, values: np.ndarray, amount: float = 0.1) -> Dict[str, float]:
        """Randomly samples a small amount of points from the array (10% by default) to recalculate the statistics.
        This introduces a 'positional instability' so that accidental mirroring or re-orderings are detected."""
        rng = np.random.default_rng(0)
        subsample = rng.choice(values, int(values.size * amount))
        return {name: self._prepare_value(operation(subsample)) for name, operation in self._STATS_OPERATIONS.items()}

    def _get_random_stats(self, raster: np.ndarray, unique_values: np.ndarray) -> List[JsonDict]:
        """First it randomly samples a few values from the list of unique values. Then for each one it checks where
        this value is located in the original array and randomly selects one of its locations. Selected locations
        and values are used for statistics."""
        rng = np.random.default_rng(0)
        randomly_chosen_values = rng.choice(
            unique_values,
            size=min(self.max_random_values, unique_values.size),
            replace=False,
        )

        random_stats: List[JsonDict] = []
        for value in randomly_chosen_values:
            value_mask = np.isnan(raster) if np.isnan(value) else raster == value
            positions = np.argwhere(value_mask)

            num_positions = positions.shape[0]
            chosen_position_index = rng.integers(num_positions)

            random_stats.append(
                {"value": self._prepare_value(value), "position": positions[chosen_position_index].tolist()}
            )

        return random_stats

    def _prepare_value(self, value: Any) -> Any:
        """Converts a value in a way that it can be compared and serialized into a JSON. It also rounds float values."""
        if not np.isscalar(value):
            return value
        if not np.isfinite(value):
            return repr(value)
        if np.issubdtype(type(value), np.integer):
            value = cast(int, value)
            return int(value)
        if np.issubdtype(type(value), bool):
            return bool(value)
        value = cast(float, value)
        return round(float(value), self.decimals)


def check_pipeline_logs(pipeline: Pipeline) -> None:
    """A utility function which checks pipeline logs and makes sure there are no failed executions"""
    if not pipeline.logging_manager.config.save_logs:
        raise ValueError("Pipeline did not save logs, this test would be useless")

    logs_folder = pipeline.logging_manager.get_pipeline_logs_folder(pipeline.current_execution_name)

    for filename in ["failed.json", "finished.json", "pipeline-report.json", "pipeline.log"]:
        path = fs.path.combine(logs_folder, filename)
        assert pipeline.storage.filesystem.isfile(path), f"File {path} is missing"

    logs_folder = pipeline.logging_manager.get_pipeline_logs_folder(pipeline.current_execution_name, full_path=True)
    failed_filename = fs.path.combine(logs_folder, "failed.json")
    assert not pipeline.eopatch_manager.load_eopatch_filenames(
        failed_filename
    ), f"Some executions failed, check {logs_folder}"

    finished_filename = os.path.join(logs_folder, "finished.json")
    assert pipeline.eopatch_manager.load_eopatch_filenames(finished_filename), "No executions finished"


def run_and_test_pipeline(
    experiment_name: str,
    *,
    config_folder: str,
    stats_folder: str,
    folder_key: Optional[str] = None,
    reset_folder: bool = True,
    save_new_stats: bool = False,
) -> None:
    """A default way of testing a pipeline

    :param experiment_name: Name of test experiment, which defines its config and stats filenames
    :param config_folder: A path to folder containing the config file
    :param stats_folder: A path to folder containing the file with expected result stats
    :param folder_key: Type of the folder containing results of the pipeline, if missing it's inferred from config
    :param reset_folder: If True it will delete content of the folder with results before running the pipeline
    :param save_new_stats: If True then new file with expected result stats will be saved, potentially overwriting the
        old one. Otherwise, the old one will be used to compare stats.
    """
    config_filename = os.path.join(config_folder, experiment_name + ".json")
    expected_stats_file = os.path.join(stats_folder, experiment_name + ".json")

    crude_configs = collect_configs_from_path(config_filename)
    raw_configs = [interpret_config_from_dict(config) for config in crude_configs]

    for index, config in enumerate(raw_configs):
        output_folder_key = folder_key or config.get("output_folder_key")
        if output_folder_key is None:
            raise ValueError(
                "Pipeline does not have a `output_folder_key` parameter, `folder_key` must be set by hand."
            )

        pipeline = load_pipeline_class(config).from_raw_config(config)

        folder = pipeline.storage.get_folder(output_folder_key)
        filesystem = pipeline.storage.filesystem

        if reset_folder:
            filesystem.removetree(folder)
        pipeline.run()

        check_pipeline_logs(pipeline)

        if index < len(raw_configs) - 1:
            continue

        tester = ContentTester(filesystem, folder)

        if save_new_stats:
            tester.save(expected_stats_file)

        stats_difference = tester.compare(expected_stats_file)
        if stats_difference:
            stats_difference_repr = stats_difference.to_json(indent=2, sort_keys=True)
            raise AssertionError(f"Expected and obtained stats differ:\n{stats_difference_repr}")


def _round_point_coords(x: float, y: float, decimals: int) -> Tuple[float, float]:
    """Rounds coordinates of a point"""
    return round(x, decimals), round(y, decimals)


def create_folder_dict(config_folder: str, stats_folder: str, subfolder: Optional[str] = None) -> Dict[str, str]:
    return {
        "config_folder": os.path.join(config_folder, subfolder) if subfolder else config_folder,
        "stats_folder": os.path.join(stats_folder, subfolder) if subfolder else config_folder,
    }


def generate_tiff_file(
    filesystem: FS,
    file_paths: Iterable[str],
    *,
    tiff_bbox: BBox,
    width: int,
    height: int,
    num_bands: int,
    dtype: type,
    seed: int = 42,
) -> None:
    """Generates tiff files containing random data."""
    transform = rasterio.transform.from_bounds(*tiff_bbox, width=width, height=height)

    generator = np.random.default_rng(seed)
    shape = (num_bands, height, width) if num_bands is not None else (height, width)

    for path in file_paths:
        with filesystem.openbin(path, "w") as file_handle:
            with rasterio.open(
                file_handle,
                "w",
                driver="GTiff",
                width=width,
                height=height,
                count=num_bands,
                dtype=dtype,
                nodata=0,
                transform=transform,
                crs=tiff_bbox.crs.ogc_string(),
            ) as dst:
                data = 10000 * generator.random(shape)
                dst.write(data)



"""
This module handles everything regarding storage of the data
"""
from typing import Dict, Literal, Optional

import fs
from fs.base import FS
from pydantic import BaseSettings, Field

from eolearn.core.utils.fs import get_aws_credentials, get_filesystem, is_s3_path
from sentinelhub import SHConfig

from ..utils.types import AwsAclType
from .base import EOGrowObject
from .schemas import ManagerSchema


class StorageManager(EOGrowObject):
    PRESET_FOLDERS: Dict[str, str] = {"logs": "logs", "input_data": "input-data", "cache": "cache"}

    class Schema(ManagerSchema, BaseSettings):
        project_folder: str = Field(
            description=(
                "The root project folder. Can be either local or on AWS S3 Bucket."
                "If on AWS, the path must be prefixed with s3://."
            ),
        )
        aws_profile: Optional[str] = Field(
            env="AWS_PROFILE",
            description=(
                "The AWS profile with credentials needed to access the S3 buckets. In case the profile isn't specified"
                " with a parameter it can be read from an environmental variable."
            ),
        )
        aws_acl: Optional[AwsAclType] = Field(
            description=(
                "An optional parameter to specify under what kind of access control list (ACL) objects should be saved"
                " to an AWS S3 bucket."
            )
        )
        structure: Dict[str, str] = Field(
            default_factory=dict,
            description="A flat key: value store mapping each key to a path in the project.",
        )
        geopandas_backend: Literal["fiona", "pyogrio"] = Field(
            "fiona", description="Which backend is used for IO operations when using geopandas."
        )

        class Config(ManagerSchema.Config):
            case_sensitive = True
            env_prefix = "eogrow_"

    config: Schema

    def __init__(self, config: Schema):
        super().__init__(config)

        for folder_key, folder_path in self.PRESET_FOLDERS.items():
            if folder_key not in self.config.structure:
                self.config.structure[folder_key] = folder_path

        self.sh_config = self._prepare_sh_config()
        self.filesystem = self._prepare_filesystem()

    def _prepare_sh_config(self) -> SHConfig:
        """Prepares an instance of `SHConfig` containing AWS credentials. In case given AWS profile doesn't exist it
        will show a warning and return a config without AWS credentials."""
        sh_config = SHConfig(hide_credentials=True)

        if self.is_on_aws() and self.config.aws_profile:
            sh_config = get_aws_credentials(aws_profile=self.config.aws_profile, config=sh_config)

        return sh_config

    def _prepare_filesystem(self) -> FS:
        """Prepares a filesystem object with the configuration parameters."""
        fs_kwargs: Dict[str, str] = {}
        if is_s3_path(self.config.project_folder) and self.config.aws_acl:
            fs_kwargs["acl"] = self.config.aws_acl

        return get_filesystem(self.config.project_folder, create=True, config=self.sh_config, **fs_kwargs)

    def get_folder(self, key: str, full_path: bool = False) -> str:
        """Returns the path associated with the given key in the structure config."""
        folder_path = self.config.structure[key]
        self.filesystem.makedirs(folder_path, recreate=True)

        if full_path:
            return fs.path.combine(self.config.project_folder, folder_path)
        return folder_path

    def get_logs_folder(self, full_path: bool = False) -> str:
        """Method for obtaining the logs folder."""
        return self.get_folder("logs", full_path=full_path)

    def get_cache_folder(self, full_path: bool = False) -> str:
        """Returns the path associated with the cache key."""
        return self.get_folder("cache", full_path=full_path)

    def get_input_data_folder(self, full_path: bool = False) -> str:
        """Returns the path associated with the input_data key."""
        return self.get_folder("input_data", full_path=full_path)

    def is_on_aws(self) -> bool:
        """Returns True if the project_folder is on S3, False otherwise."""
        return is_s3_path(self.config.project_folder)



"""
A module containing general utilities that haven't been sorted in any other module
"""
import datetime as dt
import math
from enum import Enum
from typing import Tuple, Union

import numpy as np

from sentinelhub import BBox, DataCollection
from sentinelhub.data_collections import DataCollectionDefinition


def jsonify(param: object) -> Union[str, list]:
    """Transforms an object into a normal string."""
    if isinstance(param, set):
        return list(param)

    if isinstance(param, (dt.datetime, dt.date)):
        return param.isoformat()

    if isinstance(param, DataCollectionDefinition):
        return DataCollection(param).name
    if isinstance(param, DataCollection):
        return param.name

    if isinstance(param, Enum):
        return param.value

    raise TypeError(f"Object of type {type(param)} is not yet supported in jsonify utility function")


def reduce_to_coprime(number1: int, number2: int) -> Tuple[int, int]:
    """Divides given numbers by their greatest common divisor, thus making them coprime."""
    gcd = math.gcd(number1, number2)
    return number1 // gcd, number2 // gcd


def convert_to_int(values: np.ndarray, raise_diff: bool, error: float = 1e-8) -> np.ndarray:
    """Converts an array of floats into array of integers.

    :param values: An array of float values to be converted.
    :param raise_diff: Whether to raise an error if float values differ from integer values for more than the expected
        error.
    :param error: A joined maximal expected numerical error.
    """
    rounded_values = np.round(values)

    if raise_diff:
        diff = np.amax(np.absolute(values - rounded_values))
        if diff > error:
            raise ValueError(
                "Values can't be converted to integers because the difference to nearest integer values is larger "
                "than expected numerical error"
            )

    return rounded_values.astype(int)


def convert_bbox_coords_to_int(bbox: BBox, error: float = 1e-8) -> BBox:
    """Converts bounding box coordinates to integers by removing numerical errors. If the difference is larger than a
    numerical error it raises an error."""
    coords = np.array(list(bbox))
    fixed_coords = convert_to_int(coords, raise_diff=True, error=error)
    return BBox(tuple(fixed_coords), crs=bbox.crs)


def large_list_repr(large_list: list) -> str:
    """Creates a representation of a large list of elements that consists only of a representation of first 3 and the
    last element."""
    if len(large_list) <= 4:
        return repr(large_list)

    first_elements = ", ".join(map(repr, large_list[:3]))
    return f"[{first_elements}, ..., {large_list[-1]}]"



"""
Pipelines for testing
"""
import logging
from typing import Any, Dict, List, Optional, Tuple, Type, TypeVar

import numpy as np
from pydantic import Field

from eolearn.core import CreateEOPatchTask, EONode, EOWorkflow, MergeEOPatchesTask, OverwritePermission, SaveTask

from ..core.config import RawConfig, recursive_config_join
from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema
from ..tasks.testing import DummyRasterFeatureTask, DummyTimestampFeatureTask
from ..utils.types import Feature, TimePeriod
from ..utils.validators import field_validator, parse_dtype, parse_time_period

Self = TypeVar("Self", bound="TestPipeline")
LOGGER = logging.getLogger(__name__)


class TestPipeline(Pipeline):
    """Pipeline that just tests if all managers works correctly. It can be used to check if area manager creates a
    correct grid.
    """

    class Schema(Pipeline.Schema):
        class Config:
            extra = "allow"

    _DEFAULT_CONFIG_PARAMS = {
        "pipeline": "eogrow.pipelines.testing.TestPipeline",
        "eopatch": {"manager": "eogrow.eopatches.EOPatchManager"},
        "logging": {"manager": "eogrow.logging.LoggingManager", "show_logs": True},
    }

    @classmethod
    def with_defaults(cls: Type[Self], config: RawConfig) -> Self:
        config = recursive_config_join(config, cls._DEFAULT_CONFIG_PARAMS)  # type: ignore[assignment]
        return cls.from_raw_config(config)

    def run_procedure(self) -> Tuple[List, List]:
        """Performs basic tests of managers"""
        if self.storage.filesystem.exists("/"):
            LOGGER.info("Project folder %s exists", self.storage.config.project_folder)
        else:
            LOGGER.info("Project folder %s does not exist", self.storage.config.project_folder)

        self.area_manager.get_area_dataframe()
        self.area_manager.get_area_geometry()
        grid = self.area_manager.get_grid()
        grid_size = self.area_manager.get_grid_size()
        LOGGER.info("Grid has %d EOPatches and is split over %d CRS zones", grid_size, len(grid))

        eopatches = self.eopatch_manager.get_eopatch_filenames()
        LOGGER.info("The first EOPatch has a name %s", eopatches[0])

        return [], []


class FeatureSchema(BaseSchema):
    same_for_all: bool = Field(
        False,
        description=(
            "A flag to specify if the same feature values should be generated for all EOPatches. By default each"
            " EOPatch will have different values."
        ),
    )


class RasterFeatureSchema(FeatureSchema):
    feature: Feature = Field(description="A feature to be processed.")
    shape: Tuple[int, ...] = Field(description="A shape of a feature")
    dtype: np.dtype = Field(description="The output dtype of the feature")
    _parse_dtype = field_validator("dtype", parse_dtype, pre=True)
    min_value: int = Field(0, description="All values in the feature will be greater or equal to this value.")
    max_value: int = Field(1, description="All values in the feature will be smaller to this value.")


class TimestampFeatureSchema(FeatureSchema):
    time_period: TimePeriod = Field(description="Time period from where timestamps will be generated.")
    _validate_time_period = field_validator("time_period", parse_time_period, pre=True)

    timestamp_num: int = Field(description="Number of timestamps from the interval")


class DummyDataPipeline(Pipeline):
    """Pipeline for generating dummy data."""

    class Schema(Pipeline.Schema):
        output_folder_key: str = Field(description="The storage manager key pointing to the pipeline output folder.")
        seed: Optional[int] = Field(description="A randomness seed.")

        raster_features: List[RasterFeatureSchema] = Field(
            default_factory=list, description="A list of raster features to be generated."
        )
        timestamp_feature: Optional[TimestampFeatureSchema]

    config: Schema

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self._nodes_to_configs_map: Dict[EONode, FeatureSchema] = {}

    def build_workflow(self) -> EOWorkflow:
        """Creates a workflow with tasks that generate different types of features and tasks that join and save the
        final EOPatch."""
        self._nodes_to_configs_map = {}
        start_node = EONode(CreateEOPatchTask())

        if self.config.timestamp_feature:
            timestamp_task = DummyTimestampFeatureTask(
                time_interval=self.config.timestamp_feature.time_period,
                timestamp_num=self.config.timestamp_feature.timestamp_num,
            )
            start_node = EONode(timestamp_task, inputs=[start_node])
            self._nodes_to_configs_map[start_node] = self.config.timestamp_feature

        add_feature_nodes = []
        for index, feature_config in enumerate(self.config.raster_features):
            raster_task = DummyRasterFeatureTask(
                feature_config.feature,
                shape=feature_config.shape,
                dtype=np.dtype(feature_config.dtype),
                min_value=feature_config.min_value,
                max_value=feature_config.max_value,
            )
            node = EONode(raster_task, inputs=[start_node], name=f"{DummyRasterFeatureTask.__name__}_{index}")
            add_feature_nodes.append(node)
            self._nodes_to_configs_map[node] = feature_config

        if add_feature_nodes:
            join_node = EONode(MergeEOPatchesTask(), inputs=add_feature_nodes)
            previous_node = join_node
        else:
            previous_node = start_node

        save_task = SaveTask(
            self.storage.get_folder(self.config.output_folder_key),
            filesystem=self.storage.filesystem,
            overwrite_permission=OverwritePermission.OVERWRITE_PATCH,
        )
        save_node = EONode(save_task, inputs=[previous_node])

        return EOWorkflow.from_endnodes(save_node)

    def get_execution_arguments(self, workflow: EOWorkflow) -> List[Dict[EONode, Dict[str, object]]]:
        """Extends the basic method for adding execution arguments by adding seed arguments a sampling task"""
        exec_args = super().get_execution_arguments(workflow)

        # Sorting is done to ensure seeds are always given to nodes in the same order
        add_feature_nodes = sorted(self._nodes_to_configs_map, key=lambda _node: _node.get_name())

        generator = np.random.default_rng(seed=self.config.seed)
        for index, workflow_args in enumerate(exec_args):
            for node in add_feature_nodes:
                seed: object = generator.integers(low=0, high=2**32)

                if self._nodes_to_configs_map[node].same_for_all and index > 0:
                    seed = exec_args[0][node]["seed"]

                workflow_args[node] = dict(seed=seed)

        return exec_args



"""
Pipelines for testing
"""
import logging
from typing import Any, Dict, List, Literal, Optional, Tuple

from pydantic import Field, root_validator

from eolearn.core import EONode, EOWorkflow, FeatureType, LoadTask, OverwritePermission, SaveTask
from sentinelhub import BBox

from ..core.area.base import AreaManager
from ..core.eopatch import EOPatchManager
from ..core.pipeline import Pipeline
from ..core.schemas import BaseSchema, ManagerSchema
from ..tasks.spatial import SpatialJoinTask, SpatialSliceTask
from ..utils.grid import GridTransformation
from ..utils.types import Feature, FeatureSpec, RawSchemaDict
from ..utils.validators import field_validator, optional_field_validator, validate_manager

LOGGER = logging.getLogger(__name__)


class FeatureSchema(BaseSchema):
    feature: Feature = Field(description="A feature to be processed.")
    no_data_value: float = Field(
        0,
        description=(
            "Used for spatial raster features to fill out parts of a target grid that are outside of a source grid."
        ),
    )
    unique_columns: Optional[List[str]] = Field(
        description=(
            "This is only meant to be used for vector features. If provided, it will use this list of columns to drop"
            " duplicated rows in spatially joined dataframes."
        )
    )

    @root_validator
    def check_values(cls, values: RawSchemaDict) -> RawSchemaDict:
        """Multiple different checks that given values make sense."""
        feature = values["feature"]
        feature_type = feature if isinstance(feature, FeatureType) else feature[0]

        assert feature_type is not FeatureType.BBOX, "A bounding box doesn't have to be specified in config"

        assert (
            feature is FeatureType.TIMESTAMP or feature != feature_type
        ), "All features, except timestamp, should have feature name"

        assert (
            values.get("no_data_value") == 0 or feature_type.is_raster()
        ), "Parameter 'no_data_value' can only be set for raster feature types"

        assert (
            values.get("unique_columns") is None or feature_type.is_vector()
        ), "Parameter 'unique_column' can only be set for vector feature types"

        return values


class SwitchGridsPipeline(Pipeline):
    """Pipeline that converts one EOPatch tiling grid into another.

    Notes:
      - Grid transformations are defined in area managers. A pair of grids are compatible if and only if the source
        area manager implements a transformation into the target area manager.
      - By default, if a grid transformation would cause a misalignment in any EOPatch feature, an error would be
        raised during EOWorkflow execution. Make sure to choose parameters where that won't happen.
    """

    class Schema(Pipeline.Schema):
        target_area: ManagerSchema = Field(description="An area manager configuration that defines a new grid.")
        validate_area = field_validator("target_area", validate_manager, pre=True)

        target_eopatch: Optional[ManagerSchema] = Field(
            description=(
                "An EOPatch manager configuration that defines naming in the new grid. If not provided, the old naming"
                " will be used."
            )
        )
        validate_eopatch = optional_field_validator("target_eopatch", validate_manager, pre=True)

        input_folder_key: str = Field(
            description="A storage manager key pointing to the folder from where data will be loaded."
        )
        output_folder_key: str = Field(
            description="A storage manager key pointing to the folder to where data will be saved."
        )
        features: List[FeatureSchema] = Field(description="Which features will be loaded and joined.")
        raise_misaligned: bool = Field(
            True,
            description=(
                "If True this will raise an error if splitting or joining any spatial raster EOPatch feature would "
                "cause a misalignment. If False, misalignment issues will be ignored."
            ),
        )
        patch_list: None = None
        input_patch_file: None = None
        skip_existing: Literal[False] = False

    config: Schema

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)

        self.target_area_manager: AreaManager = self._load_manager(self.config.target_area, storage=self.storage)

        target_eopatch_config = self.config.target_eopatch or self.config.eopatch
        self.target_eopatch_manager: EOPatchManager = self._load_manager(
            target_eopatch_config, area_manager=self.target_area_manager
        )

    def run_procedure(self) -> Tuple[List, List]:
        """A procedure that defines a transformation between grids and transforms EOPatches."""
        transformations = self.area_manager.transform_grid(self.target_area_manager)
        if not transformations:
            return [], []

        workflow = self.build_workflow(transformations)
        exec_args = self.get_execution_arguments_from_transformations(workflow, transformations)
        execution_names, name_mapping = self._get_execution_names_and_mapping(transformations)

        finished, failed, _ = self.run_execution(workflow, exec_args, eopatch_list=execution_names)

        finished = [input_name for execution_name in finished for input_name in name_mapping[execution_name]]
        failed = [input_name for execution_name in failed for input_name in name_mapping[execution_name]]
        return finished, failed

    def build_workflow(self, transformations: List[GridTransformation]) -> EOWorkflow:
        """Create a workflow that consists of loading, joining groups of EOPatches, slicing into new EOPatches and
        saving them."""
        features = self._get_features()

        input_path = self.storage.get_folder(self.config.input_folder_key)
        max_input_patch_num = max(len(transformation.source_bboxes) for transformation in transformations)
        load_nodes = [
            EONode(LoadTask(input_path, filesystem=self.storage.filesystem, features=features))
            for _ in range(max_input_patch_num)
        ]

        join_task = SpatialJoinTask(
            features,
            no_data_map=self._get_no_data_map(),
            unique_columns_map=self._get_unique_columns_map(),
            raise_misaligned=self.config.raise_misaligned,
        )
        join_node = EONode(join_task, inputs=load_nodes)

        save_nodes = []
        output_path = self.storage.get_folder(self.config.output_folder_key)
        max_output_patch_num = max(len(transformation.target_bboxes) for transformation in transformations)
        for _ in range(max_output_patch_num):
            slice_task = SpatialSliceTask(features, raise_misaligned=self.config.raise_misaligned)
            slice_node = EONode(slice_task, inputs=[join_node])

            save_task = SaveTask(
                output_path,
                filesystem=self.storage.filesystem,
                features=features,
                overwrite_permission=OverwritePermission.OVERWRITE_FEATURES,
            )
            save_node = EONode(save_task, inputs=[slice_node])
            save_nodes.append(save_node)

        return EOWorkflow.from_endnodes(*save_nodes)

    def get_execution_arguments_from_transformations(
        self, workflow: EOWorkflow, transformations: List[GridTransformation]
    ) -> List[Dict[EONode, Dict[str, object]]]:
        """Creates execution arguments."""
        nodes = workflow.get_nodes()
        load_nodes = [node for node in nodes if isinstance(node.task, LoadTask)]
        save_nodes = [node for node in nodes if isinstance(node.task, SaveTask)]
        join_node = [node for node in nodes if isinstance(node.task, SpatialJoinTask)][0]
        slice_nodes = [save_node.inputs[0] for save_node in save_nodes]

        return [
            {
                join_node: dict(bbox=transformation.enclosing_bbox),
                **self._get_io_exec_dict(transformation, load_nodes=load_nodes, save_nodes=save_nodes),
                **self._get_slice_exec_dict(transformation, slice_nodes),
            }
            for transformation in transformations
        ]

    def _get_io_exec_dict(
        self, transformation: GridTransformation, *, load_nodes: List[EONode], save_nodes: List[EONode]
    ) -> Dict[EONode, Dict[str, Any]]:
        """Prepares execution arguments for load and save nodes."""
        input_names = list(self.eopatch_manager.generate_names(transformation.source_df))
        input_names += [None] * (len(load_nodes) - len(input_names))

        output_names = list(self.target_eopatch_manager.generate_names(transformation.target_df))
        output_names += [None] * (len(save_nodes) - len(output_names))

        return {
            node: dict(eopatch_folder=name) for name, node in zip(input_names + output_names, load_nodes + save_nodes)
        }

    @staticmethod
    def _get_slice_exec_dict(
        transformation: GridTransformation, slice_nodes: List[EONode]
    ) -> Dict[EONode, Dict[str, Any]]:
        """Prepares execution arguments for slice task nodes."""
        slice_bboxes: List[Optional[BBox]] = list(transformation.target_bboxes)
        slice_bboxes += [None] * (len(slice_nodes) - len(transformation.target_bboxes))

        return {node: dict(bbox=bbox) for bbox, node in zip(slice_bboxes, slice_nodes)}

    def _get_execution_names_and_mapping(
        self, transformations: List[GridTransformation]
    ) -> Tuple[List[str], Dict[str, List[str]]]:
        """An execution name will always be the name of alphabetically first name of input EOPatches. That is because
        each output EOPatch should be in exactly 1 transformation while an input EOPatch can be used in multiple
        transformations. A mapping will be provided from an execution name to a list of input EOPatches."""
        names: List[str] = []
        mapping: Dict[str, List[str]] = {}

        for transformation in transformations:
            input_names = list(self.eopatch_manager.generate_names(transformation.source_df))
            output_names = list(self.target_eopatch_manager.generate_names(transformation.target_df))
            execution_name = min(output_names)

            names.append(execution_name)
            if execution_name in mapping:
                raise ValueError(f"An output EOPatch {execution_name} is being used in more than 1 transformation.")
            mapping[execution_name] = input_names

        return names, mapping

    def _get_features(self) -> List[FeatureSpec]:
        """Provides features that will be transformed by the pipeline."""
        features = [feature_config.feature for feature_config in self.config.features]
        meta_features: List[FeatureSpec] = [FeatureType.BBOX]
        if any(f_type.is_temporal() for f_type, _ in features):
            meta_features += [FeatureType.TIMESTAMP]

        return meta_features + features  # type: ignore[operator]

    def _get_no_data_map(self) -> Dict[Feature, float]:
        """Provides a map between spatial raster features and their 'no data' values."""
        no_data_map: Dict[Feature, float] = {}
        for feature_config in self.config.features:
            feature = feature_config.feature
            feature_type = feature if isinstance(feature, FeatureType) else feature[0]

            if feature_type.is_spatial() and feature_type.is_raster():
                no_data_map[feature] = feature_config.no_data_value

        return no_data_map

    def _get_unique_columns_map(self) -> Dict[Feature, List[str]]:
        """Provides a map between vector raster features and their unique columns."""
        unique_columns_map: Dict[Feature, List[str]] = {}
        for feature_config in self.config.features:
            if feature_config.unique_columns:
                unique_columns_map[feature_config.feature] = feature_config.unique_columns
        return unique_columns_map



"""
Task definitions for prediction
"""
import abc
from typing import Any, Callable, List, Optional, Tuple, Union, cast

import fs
import joblib
import numpy as np
from fs.base import FS

from eolearn.core import EOPatch, EOTask, execute_with_mp_lock
from eolearn.core.utils.fs import pickle_fs, unpickle_fs

from ..utils.types import Feature


class BasePredictionTask(EOTask, metaclass=abc.ABCMeta):
    """Base predictions task streamlining data preprocessing before prediction"""

    def __init__(
        self,
        *,
        model_path: str,
        filesystem: FS,
        input_features: List[Feature],
        mask_feature: Feature,
        output_feature: Feature,
        output_dtype: Optional[np.dtype],
        mp_lock: bool,
    ):
        """
        :param model_path: A file path to the model. The path is relative to the filesystem object.
        :param filesystem: A filesystem object.
        :param input_features: List of features containing input for the model, which are concatenated in given order
        :param mask_feature: Mask specifying which points are to be predicted
        :param output_feature: Feature into which predictions are saved
        :param mp_lock: If predictions should be executed with a multiprocessing lock
        """
        self.model_path = model_path
        self._model = None
        self.pickled_filesystem = pickle_fs(filesystem)

        self.input_features = input_features
        self.mask_feature = mask_feature
        self.output_feature = output_feature
        self.output_dtype = output_dtype

        self.mp_lock = mp_lock

    def process_data(self, eopatch: EOPatch, mask: np.ndarray) -> np.ndarray:
        """Masks and reshapes data into a form suitable for the model"""
        all_features = []
        for ftype, fname in self.input_features:
            array = eopatch[ftype, fname]

            if ftype.is_timeless():
                all_features.append(array[mask, :])
            else:
                array = array[:, mask, :]
                time, pixels, depth = array.shape
                array = np.moveaxis(array, 0, 1)
                all_features.append(array.reshape(pixels, time * depth))

        return np.concatenate(all_features, axis=-1)

    @property
    def model(self) -> Any:
        """Implements lazy loading that gets around filesystem issues"""
        if self._model is None:
            filesystem = unpickle_fs(self.pickled_filesystem)

            with filesystem.openbin(self.model_path, "r") as file_handle:
                self._model = joblib.load(file_handle)

        return self._model

    def apply_predictor(
        self, predictor: Callable, processed_features: np.ndarray, return_on_empty: Optional[np.ndarray] = None
    ) -> np.ndarray:
        """Helper function that applies the predictor according to the mp_lock settings"""
        if processed_features.shape[0] == 0 and return_on_empty is not None:
            return return_on_empty

        if self.mp_lock:
            predictions = execute_with_mp_lock(predictor, processed_features)
        else:
            predictions = predictor(processed_features)
        predictions = cast(np.ndarray, predictions)

        return predictions.astype(self.output_dtype) if self.output_dtype else predictions

    @abc.abstractmethod
    def add_predictions(self, eopatch: EOPatch, processed_features: np.ndarray, mask: np.ndarray) -> EOPatch:
        """Runs the model prediction on given features and adds them to the eopatch. Must reverse mask beforehand."""

    @staticmethod
    def transform_to_feature_form(
        predictions: np.ndarray, mask: np.ndarray, no_value: Union[float, int] = 0
    ) -> np.ndarray:
        """Transforms an array of predictions into an EOPatch suitable array, making sure to reverse the masking"""
        full_predictions = np.full((*mask.shape, predictions.shape[-1]), dtype=predictions.dtype, fill_value=no_value)
        full_predictions[mask, :] = predictions
        return full_predictions

    def execute(self, eopatch: EOPatch) -> EOPatch:
        """Run model on input features and save predictions to eopatch"""

        some_feature = self.input_features[0]
        mask_size = eopatch.get_spatial_dimension(*some_feature)
        mask = np.squeeze(eopatch[self.mask_feature], axis=-1) if self.mask_feature else np.ones(mask_size)
        mask = mask.astype(bool)

        preprocessed_features = self.process_data(eopatch, mask)
        eopatch = self.add_predictions(eopatch, preprocessed_features, mask)
        return eopatch


class ClassificationPredictionTask(BasePredictionTask):
    """Uses a classification model to produce predictions for given input features"""

    def __init__(
        self,
        *,
        label_encoder_filename: Optional[str],
        output_probability_feature: Optional[Feature] = None,
        **kwargs: Any,
    ):
        """
        :param label_encoder_filename: Name of file containing the label encoder with which to decode predictions. The
            file should be in the same folder as the model.
        :param output_probability_feature: If specified saves pseudo-probabilities into given feature.
        :param kwargs: Parameters of `BasePredictionTask`
        """
        self.label_encoder_filename = label_encoder_filename
        self._label_encoder = None
        self.output_probability_feature = output_probability_feature
        super().__init__(**kwargs)

    @property
    def label_encoder(self) -> Any:
        """Implements lazy loading that gets around filesystem issues"""
        if self._label_encoder is None and self.label_encoder_filename is not None:
            filesystem = unpickle_fs(self.pickled_filesystem)

            model_folder = fs.path.dirname(self.model_path)
            label_encoder_path = fs.path.join(model_folder, self.label_encoder_filename)

            with filesystem.openbin(label_encoder_path, "r") as file_handle:
                self._label_encoder = joblib.load(file_handle)

        return self._label_encoder

    def add_predictions(self, eopatch: EOPatch, processed_features: np.ndarray, mask: np.ndarray) -> EOPatch:
        """Runs the model prediction on given features and adds them to the eopatch

        If specified also adds probability scores and uses a label encoder.
        """
        predictions = self.apply_predictor(self.model.predict, processed_features, np.zeros((0,), dtype=np.uint8))
        predictions = predictions[..., np.newaxis]
        if self.label_encoder is not None:
            predictions = self.label_encoder.inverse_transform(predictions)
        eopatch[self.output_feature] = self.transform_to_feature_form(predictions, mask)

        if self.output_probability_feature is not None:
            probabilities = self.apply_predictor(self.model.predict_proba, processed_features)
            eopatch[self.output_probability_feature] = self.transform_to_feature_form(probabilities, mask)

        return eopatch


class RegressionPredictionTask(BasePredictionTask):
    """Computes values and scores given an input model and eopatch feature name"""

    def __init__(
        self,
        *,
        clip_predictions: Optional[Tuple[float, float]],
        **kwargs: Any,
    ):
        """
        :param clip_predictions: If given the task also clips predictions to the specified interval.
        :param kwargs: Parameters of `BasePredictionTask`
        """
        self.clip_predictions = clip_predictions
        super().__init__(**kwargs)

    def add_predictions(self, eopatch: EOPatch, processed_features: np.ndarray, mask: np.ndarray) -> EOPatch:
        """Runs the model prediction on given features and adds them to the eopatch. Must reverse mask beforehand."""

        predictions = self.apply_predictor(self.model.predict, processed_features, np.zeros((0,), dtype=np.float32))
        predictions = predictions[..., np.newaxis]
        if self.clip_predictions is not None:
            predictions = predictions.clip(*self.clip_predictions)
        eopatch[self.output_feature] = self.transform_to_feature_form(predictions, mask)

        return eopatch
